#!/usr/bin/env python

import re
import subprocess
from datetime import datetime

import toml

pyproject = toml.load("pyproject.toml")
version = pyproject["project"]["version"]


def get_changelog(version):
    s = open("CHANGELOG.md").read()
    search_heading = f"# Release {version}"
    curdate = datetime.today().date().isoformat()

    headings = re.findall(r"# Release \d+\.\d+\.\d+", s)
    version_heading_index = headings.index(search_heading)
    version_changelog = s[
        s.index(headings[version_heading_index]) : s.index(
            headings[version_heading_index + 1]
        )
    ].strip()

    version_changelog = insert_date(version_changelog, curdate, after=search_heading)
    version_changelog = remove_headings(version_changelog)

    return version_changelog


def remove_headings(version_changelog):
    # Remove Markdown-style headings as it matches Git comment syntax
    version_changelog = re.sub(r"^#+\s+", "", version_changelog, flags=re.MULTILINE)
    return version_changelog


def insert_date(version_changelog, curdate, after):
    # Add a date section, if it doesn't already exist in the changelog section
    if "Date:" not in version_changelog:
        version_changelog = version_changelog.replace(
            after, f"{after}\n\nDate: {curdate}", 1
        )
    return version_changelog


message = get_changelog(version)
proc = subprocess.Popen(["git", "tag", "-s", version, "-m", message])
proc.wait()

subprocess.check_call(["git", "show", version])

print()
if input("Push [yn]? ") in "yY":
    subprocess.check_output(["git", "push", "origin", version])



import pathlib

import pytest

from ropetest import testutils


@pytest.fixture
def mod1(project):
    mod1 = testutils.create_module(project, "mod1")
    return mod1


@pytest.fixture
def mod1_path(mod1):
    return pathlib.Path(mod1.real_path)


@pytest.fixture
def typing_path():
    import typing

    return pathlib.Path(typing.__file__)


@pytest.fixture
def example_external_package_module_path(
    session_venv,
    external_fixturepkg,
    session_venv_site_packages,
):
    return session_venv_site_packages / "external_fixturepkg/mod1.py"


@pytest.fixture
def example_external_package_path(
    session_venv,
    external_fixturepkg,
    session_venv_site_packages,
):
    return session_venv_site_packages / "external_fixturepkg"


@pytest.fixture
def compiled_lib():
    import _sqlite3

    return "_sqlite3", pathlib.Path(_sqlite3.__file__)



"""Project file system commands.

This modules implements file system operations used by rope.  Different
version control systems can be supported by implementing the interface
provided by `FileSystemCommands` class.  See `SubversionCommands` and
`MercurialCommands` for example.

"""

import os
import re
import shutil
import subprocess
import typing

FileContent = typing.NewType("FileContent", bytes)


def create_fscommands(root):
    dirlist = os.listdir(root)
    commands = {
        ".hg": MercurialCommands,
        ".svn": SubversionCommands,
        ".git": GITCommands,
        "_svn": SubversionCommands,
        "_darcs": DarcsCommands,
    }
    for key in commands:
        if key in dirlist:
            try:
                return commands[key](root)
            except (ImportError, OSError):
                pass
    return FileSystemCommands()


class FileSystemCommands:
    def create_file(self, path):
        open(path, "w").close()

    def create_folder(self, path):
        os.mkdir(path)

    def move(self, path, new_location):
        shutil.move(path, new_location)

    def remove(self, path):
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path)

    def write(self, path, data):
        file_ = open(path, "wb")
        try:
            file_.write(data)
        finally:
            file_.close()

    def read(self, path):
        with open(path, "rb") as handle:
            return handle.read()


class SubversionCommands:
    def __init__(self, *args):
        self.normal_actions = FileSystemCommands()
        import pysvn  # type:ignore

        self.client = pysvn.Client()

    def create_file(self, path):
        self.normal_actions.create_file(path)
        self.client.add(path, force=True)

    def create_folder(self, path):
        self.normal_actions.create_folder(path)
        self.client.add(path, force=True)

    def move(self, path, new_location):
        self.client.move(path, new_location, force=True)

    def remove(self, path):
        self.client.remove(path, force=True)

    def write(self, path, data):
        self.normal_actions.write(path, data)

    def read(self, path):
        return self.normal_actions.read(path)


class MercurialCommands:
    def __init__(self, root):
        self.hg = self._import_mercurial()
        self.normal_actions = FileSystemCommands()
        try:
            self.ui = self.hg.ui.ui(
                verbose=False,
                debug=False,
                quiet=True,
                interactive=False,
                traceback=False,
                report_untrusted=False,
            )
        except Exception:
            self.ui = self.hg.ui.ui()
            self.ui.setconfig("ui", "interactive", "no")
            self.ui.setconfig("ui", "debug", "no")
            self.ui.setconfig("ui", "traceback", "no")
            self.ui.setconfig("ui", "verbose", "no")
            self.ui.setconfig("ui", "report_untrusted", "no")
            self.ui.setconfig("ui", "quiet", "yes")

        self.repo = self.hg.hg.repository(self.ui, root)

    def _import_mercurial(self):
        import mercurial.commands  # type:ignore
        import mercurial.hg  # type:ignore
        import mercurial.ui  # type:ignore

        return mercurial

    def create_file(self, path):
        self.normal_actions.create_file(path)
        self.hg.commands.add(self.ui, self.repo, path)

    def create_folder(self, path):
        self.normal_actions.create_folder(path)

    def move(self, path, new_location):
        self.hg.commands.rename(self.ui, self.repo, path, new_location, after=False)

    def remove(self, path):
        self.hg.commands.remove(self.ui, self.repo, path)

    def write(self, path, data):
        self.normal_actions.write(path, data)

    def read(self, path):
        return self.normal_actions.read(path)


class GITCommands:
    def __init__(self, root):
        self.root = root
        self._do(["version"])
        self.normal_actions = FileSystemCommands()

    def create_file(self, path):
        self.normal_actions.create_file(path)
        self._do(["add", self._in_dir(path)])

    def create_folder(self, path):
        self.normal_actions.create_folder(path)

    def move(self, path, new_location):
        self._do(["mv", self._in_dir(path), self._in_dir(new_location)])

    def remove(self, path):
        self._do(["rm", self._in_dir(path)])

    def write(self, path, data):
        # XXX: should we use ``git add``?
        self.normal_actions.write(path, data)

    def read(self, path):
        return self.normal_actions.read(path)

    def _do(self, args):
        _execute(["git"] + args, cwd=self.root)

    def _in_dir(self, path):
        if path.startswith(self.root):
            return path[len(self.root) + 1 :]
        return self.root


class DarcsCommands:
    def __init__(self, root):
        self.root = root
        self.normal_actions = FileSystemCommands()

    def create_file(self, path):
        self.normal_actions.create_file(path)
        self._do(["add", path])

    def create_folder(self, path):
        self.normal_actions.create_folder(path)
        self._do(["add", path])

    def move(self, path, new_location):
        self._do(["mv", path, new_location])

    def remove(self, path):
        self.normal_actions.remove(path)

    def read(self, path):
        return self.normal_actions.read(path)

    def write(self, path, data):
        self.normal_actions.write(path, data)

    def _do(self, args):
        _execute(["darcs"] + args, cwd=self.root)


def _execute(args, cwd=None):
    process = subprocess.Popen(args, cwd=cwd, stdout=subprocess.PIPE)
    process.wait()
    return process.returncode


def unicode_to_file_data(contents: str, encoding=None, newlines=None) -> FileContent:
    assert isinstance(contents, str)
    if newlines and newlines != "\n":
        contents = contents.replace("\n", newlines)
    if encoding is None:
        encoding = read_str_coding(contents)
    if encoding is not None:
        return FileContent(contents.encode(encoding))
    try:
        return FileContent(contents.encode())
    except UnicodeEncodeError:
        return FileContent(contents.encode("utf-8"))


def file_data_to_unicode(data, encoding=None):
    result = _decode_data(data, encoding)
    newline = "\n"
    if "\r\n" in result:
        result = result.replace("\r\n", "\n")
        newline = "\r\n"
    if "\r" in result:
        result = result.replace("\r", "\n")
        newline = "\r"
    return result, newline


def _decode_data(data, encoding):
    if isinstance(data, str):
        return data
    if encoding is None:
        encoding = read_str_coding(data)
    if encoding is None:
        # there is no encoding tip, we need to guess.
        # PEP263 says that "encoding not explicitly defined" means it is ascii,
        # but we will use utf8 instead since utf8 fully covers ascii and btw is
        # the only non-latin sane encoding.
        encoding = "utf-8"
    try:
        return data.decode(encoding)
    except (UnicodeError, LookupError):
        # fallback to latin1: it should never fail
        return data.decode("latin1")


def read_str_coding(source):
    # as defined by PEP-263 (https://www.python.org/dev/peps/pep-0263/)
    CODING_LINE_PATTERN = b"^[ \t\f]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)"

    if type(source) == bytes:
        newline = b"\n"
        CODING_LINE_PATTERN = re.compile(CODING_LINE_PATTERN)
    else:
        newline = "\n"
        CODING_LINE_PATTERN = re.compile(CODING_LINE_PATTERN.decode("ascii"))
    for line in source.split(newline, 2)[:2]:
        if re.match(CODING_LINE_PATTERN, line):
            return _find_coding(line)
    else:
        return


def _find_coding(text):
    if isinstance(text, str):
        text = text.encode("utf-8")
    coding = b"coding"
    to_chr = chr
    try:
        start = text.index(coding) + len(coding)
        if text[start] not in b"=:":
            return
        start += 1
        while start < len(text) and to_chr(text[start]).isspace():
            start += 1
        end = start
        while end < len(text):
            c = text[end]
            if not to_chr(c).isalnum() and c not in b"-_":
                break
            end += 1
        result = text[start:end]
        if isinstance(result, bytes):
            result = result.decode("utf-8")
        return result
    except ValueError:
        pass



from __future__ import annotations

import unittest
from textwrap import dedent
from typing import TYPE_CHECKING, Union

from rope.base import exceptions
from rope.refactor import move
from ropetest import testutils


if TYPE_CHECKING:
    from rope.base import resources, project

class MoveRefactoringTest(unittest.TestCase):
    project: project.Project
    mod1: resources.File
    mod2: resources.File
    mod3: resources.File
    pkg: resources.Folder
    mod4: resources.File
    mod5: resources.File
    origin_module: resources.File
    destination_module: resources.File
    origin_module_in_pkg: resources.File
    destination_module_in_pkg: resources.File
    destination_pkg_root: resources.Folder

    def setUp(self) -> None:
        super().setUp()
        self.project = testutils.sample_project()
        self.mod1 = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")
        self.mod3 = testutils.create_module(self.project, "mod3")
        self.pkg = testutils.create_package(self.project, "pkg")
        self.mod4 = testutils.create_module(self.project, "mod4", self.pkg)
        self.mod5 = testutils.create_module(self.project, "mod5", self.pkg)
        self.origin_module = testutils.create_module(self.project, "origin_module")
        self.destination_module = testutils.create_module(self.project, "destination_module")
        self.origin_module_in_pkg = testutils.create_module(self.project, "origin_module_in_pkg", self.pkg)
        self.destination_module_in_pkg = testutils.create_module(self.project, "destination_module_in_pkg", self.pkg)
        self.destination_pkg_root = testutils.create_package(self.project, "destination_pkg_root")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _move(
        self,
        resource: Union[resources.File, resources.Folder],
        offset: Union[int, None],
        dest_resource: Union[str, resources.File, resources.Folder],
    ):
        mover = move.create_move(self.project, resource, offset)
        changes = mover.get_changes(dest_resource)
        self.project.do(changes)

    def _move_to_attr(
        self,
        resource: Union[resources.File, resources.Folder],
        offset: Union[int, None],
        dest_attr: str,
        *,
        new_name: str,
    ):
        mover = move.create_move(self.project, resource, offset)
        assert isinstance(mover, move.MoveMethod)
        changes = mover.get_changes(dest_attr, new_name=new_name)
        self.project.do(changes)

    def test_move_constant(self) -> None:
        self.origin_module.write("foo = 123\n")
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, self.destination_module)
        self.assertEqual("", self.origin_module.read())
        self.assertEqual("foo = 123\n", self.destination_module.read())

    def test_move_constant_2(self) -> None:
        self.origin_module.write("bar = 321\nfoo = 123\n")
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, self.destination_module)
        self.assertEqual("bar = 321\n", self.origin_module.read())
        self.assertEqual("foo = 123\n", self.destination_module.read())

    def test_move_target_is_module_name(self) -> None:
        self.origin_module.write("foo = 123\n")
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, "destination_module")
        self.assertEqual("", self.origin_module.read())
        self.assertEqual("foo = 123\n", self.destination_module.read())

    def test_move_target_is_package_name(self) -> None:
        self.origin_module.write("foo = 123\n")
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, "pkg.destination_module_in_pkg")
        self.assertEqual("", self.origin_module.read())
        self.assertEqual("foo = 123\n", self.destination_module_in_pkg.read())

    def test_move_constant_multiline(self) -> None:
        self.origin_module.write(dedent("""\
            foo = (
                123
            )
        """))
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, self.destination_module)
        self.assertEqual("", self.origin_module.read())
        self.assertEqual(
            dedent("""\
                foo = (
                    123
                )
            """),
            self.destination_module.read(),
        )

    def test_move_constant_multiple_statements(self) -> None:
        self.origin_module.write(dedent("""\
            foo = 123
            foo += 3
            foo = 4
        """))
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import destination_module
                destination_module.foo += 3
                destination_module.foo = 4
            """),
            self.origin_module.read(),
        )
        self.assertEqual("foo = 123\n", self.destination_module.read())

    def test_simple_moving(self) -> None:
        """Move a global class definition"""
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual("", self.origin_module.read())
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    pass
            """),
            self.destination_module.read(),
        )

    def test_moving_with_comment_prefix(self) -> None:
        """Comments above the moved class are moved to the destination module"""
        self.origin_module.write(dedent("""\
            a = 1
            # 1
            # 2
            class AClass(object):
                pass
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual("a = 1\n", self.origin_module.read())
        self.assertEqual(
            dedent("""\
                # 1
                # 2
                class AClass(object):
                    pass
            """),
            self.destination_module.read(),
        )

    def test_moving_with_comment_prefix_imports(self) -> None:
        self.origin_module.write(dedent("""\
            import foo
            a = 1
            # 1
            # 2
            class AClass(foo.FooClass):
                pass
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual("a = 1\n", self.origin_module.read())
        self.assertEqual(
            dedent("""\
                import foo


                # 1
                # 2
                class AClass(foo.FooClass):
                    pass
            """),
            self.destination_module.read(),
        )

    def test_changing_other_modules_replacing_normal_imports(self) -> None:
        """
        When moving a class from origin_module to destination_module,
        references to the class in mod3 is updated to point to
        destination_module
        """
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import destination_module
                a_var = destination_module.AClass()
            """),
            self.mod3.read(),
        )

    def test_changing_other_modules_adding_normal_imports(self) -> None:
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import origin_module
                import destination_module
                a_var = destination_module.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_prefer_from_module(self) -> None:
        self.project.prefs["prefer_module_from_imports"] = True
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        # Move to destination_module_in_pkg which is in a different package
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module_in_pkg)
        self.assertEqual(
            dedent("""\
                import origin_module
                from pkg import destination_module_in_pkg
                a_var = destination_module_in_pkg.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_preferred_import_style_is_normal_import(self) -> None:
        self.project.prefs.imports.preferred_import_style = "normal-import"
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        # Move to destination_module_in_pkg which is in a different package
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module_in_pkg)
        self.assertEqual(
            dedent("""\
                import origin_module
                import pkg.destination_module_in_pkg
                a_var = pkg.destination_module_in_pkg.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_preferred_import_style_is_from_module(self) -> None:
        self.project.prefs.imports.preferred_import_style = "from-module"
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        # Move to destination_module_in_pkg which is in a different package
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module_in_pkg)
        self.assertEqual(
            dedent("""\
                import origin_module
                from pkg import destination_module_in_pkg
                a_var = destination_module_in_pkg.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_preferred_import_style_is_from_global(self) -> None:
        self.project.prefs.imports.preferred_import_style = "from-global"
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        # Move to destination_module_in_pkg which is in a different package
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module_in_pkg)
        self.assertEqual(
            dedent("""\
                import origin_module
                from pkg.destination_module_in_pkg import AClass
                a_var = AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_noprefer_from_module(self) -> None:
        self.project.prefs["prefer_module_from_imports"] = False
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        # Move to destination_module_in_pkg which is in a different package
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module_in_pkg)
        self.assertEqual(
            dedent("""\
                import origin_module
                import pkg.destination_module_in_pkg
                a_var = pkg.destination_module_in_pkg.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_prefer_from_module_top_level_module(self) -> None:
        self.project.prefs["prefer_module_from_imports"] = True
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import origin_module
            a_var = origin_module.AClass()
            origin_module.a_function()"""))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import origin_module
                import destination_module
                a_var = destination_module.AClass()
                origin_module.a_function()"""),
            self.mod3.read(),
        )

    def test_changing_other_modules_removing_from_imports(self) -> None:
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
        """))
        self.mod3.write(dedent("""\
            from origin_module import AClass
            a_var = AClass()
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import destination_module
                a_var = destination_module.AClass()
            """),
            self.mod3.read(),
        )

    def test_changing_source_module(self) -> None:
        """
        Add import statements to the source module as the moved class now lives
        in destination_module.
        """
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
            a_var = AClass()
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                import destination_module
                a_var = destination_module.AClass()
            """),
            self.origin_module.read(),
        )

    def test_changing_destination_module(self) -> None:
        """
        Remove import statements in the destination module as the moved class
        can now be referenced from destination_module without import.
        """
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
        """))
        self.destination_module.write(dedent("""\
            from origin_module import AClass
            a_var = AClass()
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    pass
                a_var = AClass()
            """),
            self.destination_module.read(),
        )

    def test_folder_destination(self) -> None:
        folder = self.project.root.create_folder("folder")
        self.origin_module.write(dedent("""\
            class AClass(object):
                pass
        """))
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Move destination for non-modules should not be folders\.",
        ) as e:
            self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, folder)

    def test_raising_exception_for_moving_non_global_elements(self) -> None:
        self.origin_module.write(dedent("""\
            def a_func():
                class AClass(object):
                    pass
        """))
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Move only works on global classes/functions/variables, modules and methods\.",
        ):
            self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)

    def test_raising_an_exception_for_moving_non_global_variable(self) -> None:
        code = dedent("""\
            class TestClass:
                CONSTANT = 5
        """)
        self.origin_module.write(code)
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            "Move refactoring should be performed on a global class, function or variable\.",
        ):
            mover = move.create_move(
                self.project, self.origin_module, code.index("CONSTANT") + 1
            )

    def test_raising_exception_for_moving_glob_elements_to_the_same_module(self) -> None:
        self.origin_module.write(dedent("""\
            def a_func():
                pass
        """))
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            "Moving global elements to the same module\.",
        ):
            self._move(self.origin_module, self.origin_module.read().index("a_func"), self.origin_module)

    def test_moving_used_imports_to_destination_module(self) -> None:
        """
        Add import statements for imported references used by the moved
        function to the destination module.
        """
        self.mod3.write("a_var = 10")
        code = dedent("""\
            import mod3
            from mod3 import a_var
            def a_func():
                print(mod3, a_var)
        """)
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("a_func") + 1, self.destination_module)
        expected = dedent("""\
            import mod3
            from mod3 import a_var


            def a_func():
                print(mod3, a_var)
        """)
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_used_names_to_destination_module2(self) -> None:
        """
        Add import statements for references to globals in the source module
        used by the moved function to the destination module.
        """
        code = dedent("""\
            a_var = 10
            def a_func():
                print(a_var)
        """)
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("a_func") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                a_var = 10
            """),
            self.origin_module.read(),
        )
        expected = dedent("""\
            from origin_module import a_var


            def a_func():
                print(a_var)
        """)
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_used_underlined_names_to_destination_module(self) -> None:
        code = dedent("""\
            _var = 10
            def a_func():
                print(_var)
        """)
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("a_func") + 1, self.destination_module)
        expected = dedent("""\
            from origin_module import _var


            def a_func():
                print(_var)
        """)
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_and_used_relative_imports(self) -> None:
        """Move global function where the source module is in a package"""
        code = dedent("""\
            import mod5
            def a_func():
                print(mod5)
        """)
        self.origin_module_in_pkg.write(code)
        self._move(self.origin_module_in_pkg, code.index("a_func") + 1, self.destination_module)
        expected = dedent("""\
            import pkg.mod5


            def a_func():
                print(pkg.mod5)
        """)
        self.assertEqual(expected, self.destination_module.read())
        self.assertEqual("", self.origin_module_in_pkg.read())

    def test_moving_modules_into_package(self) -> None:
        """Move global function where the destination module is in a package"""
        code = dedent("""\
            import mod1
            print(mod1)"""
        )
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("mod1") + 1, self.destination_pkg_root)
        expected = dedent("""\
            import destination_pkg_root.mod1
            print(destination_pkg_root.mod1)"""
        )
        self.assertEqual(expected, self.origin_module.read())
        self.assertTrue(
            not self.mod1.exists() and self.project.find_module("destination_pkg_root.mod1") is not None
        )

    def test_moving_modules_and_removing_out_of_date_imports(self) -> None:
        code = dedent("""\
            import pkg.mod4
            print(pkg.mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("mod4") + 1, self.project.root)
        expected = dedent("""\
            import mod4
            print(mod4)""")
        self.assertEqual(expected, self.origin_module.read())
        self.assertTrue(self.project.find_module("mod4") is not None)

    def test_moving_modules_and_removing_out_of_date_froms(self) -> None:
        code = dedent("""\
            from pkg import mod4
            print(mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("mod4") + 1, self.project.root)
        self.assertEqual(
            dedent("""\
                import mod4
                print(mod4)"""
            ), 
            self.origin_module.read(),
        )

    def test_moving_modules_and_removing_out_of_date_froms2(self) -> None:
        self.mod4.write("a_var = 10")
        code = dedent("""\
            from pkg.mod4 import a_var
            print(a_var)
        """)
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("mod4") + 1, self.project.root)
        expected = dedent("""\
            from mod4 import a_var
            print(a_var)
        """)
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_modules_and_relative_import(self) -> None:
        self.mod4.write(dedent("""\
            import mod5
            print(mod5)
        """))
        code = dedent("""\
            import pkg.mod4
            print(pkg.mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("mod4") + 1, self.project.root)
        moved = self.project.find_module("mod4")
        assert moved
        expected = dedent("""\
            import pkg.mod5
            print(pkg.mod5)
        """)
        self.assertEqual(expected, moved.read())

    def test_moving_module_kwarg_same_name_as_old(self) -> None:
        self.origin_module.write(dedent("""\
            def foo(origin_module=0):
                pass"""))
        code = dedent("""\
            import origin_module
            origin_module.foo(origin_module=1)""")
        self.mod2.write(code)
        self._move(self.origin_module, None, self.destination_pkg_root)
        moved = self.project.find_module("mod2")
        assert moved
        expected = dedent("""\
            import destination_pkg_root.origin_module
            destination_pkg_root.origin_module.foo(origin_module=1)""")
        self.assertEqual(expected, moved.read())

    def test_moving_packages(self) -> None:
        code = dedent("""\
            import pkg.mod4
            print(pkg.mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("pkg") + 1, self.destination_pkg_root)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod5") is not None)
        expected = dedent("""\
            import destination_pkg_root.pkg.mod4
            print(destination_pkg_root.pkg.mod4)""")
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_modules_with_self_imports(self) -> None:
        self.mod1.write(dedent("""\
            import mod1
            print(mod1)
        """))
        self.origin_module.write(dedent("""\
            import mod1
        """))
        self._move(self.origin_module, self.origin_module.read().index("mod1") + 1, self.destination_pkg_root)
        moved = self.project.find_module("destination_pkg_root.mod1")
        assert moved
        self.assertEqual(
            dedent("""\
                import destination_pkg_root.mod1
                print(destination_pkg_root.mod1)
            """),
            moved.read(),
        )

    def test_moving_modules_with_from_imports(self) -> None:
        code = dedent("""\
            from pkg import mod4
            print(mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("pkg") + 1, self.destination_pkg_root)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod5") is not None)
        expected = dedent("""\
            from destination_pkg_root.pkg import mod4
            print(mod4)""")
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_modules_with_from_import(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import origin_module_in_pkg
            print(origin_module_in_pkg)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            from pkg2.pkg3.pkg4 import origin_module_in_pkg
            print(origin_module_in_pkg)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_multi_from_imports(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import origin_module_in_pkg, mod5
            print(origin_module_in_pkg)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            from pkg import mod5
            from pkg2.pkg3.pkg4 import origin_module_in_pkg
            print(origin_module_in_pkg)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_from_and_normal_imports(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import origin_module_in_pkg
            import pkg.origin_module_in_pkg
            print(origin_module_in_pkg)
            print(pkg.origin_module_in_pkg)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.origin_module_in_pkg
            from pkg2.pkg3.pkg4 import origin_module_in_pkg
            print(origin_module_in_pkg)
            print(pkg2.pkg3.pkg4.origin_module_in_pkg)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_normal_and_from_imports(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            import pkg.origin_module_in_pkg
            from pkg import origin_module_in_pkg
            print(origin_module_in_pkg)
            print(pkg.origin_module_in_pkg)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.origin_module_in_pkg
            from pkg2.pkg3.pkg4 import origin_module_in_pkg
            print(origin_module_in_pkg)
            print(pkg2.pkg3.pkg4.origin_module_in_pkg)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_from_import_variable(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg.origin_module_in_pkg import foo
            print(foo)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            from pkg2.pkg3.pkg4.origin_module_in_pkg import foo
            print(foo)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_normal_import(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            import pkg.origin_module_in_pkg
            print(pkg.origin_module_in_pkg)""")
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.origin_module_in_pkg") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.origin_module_in_pkg
            print(pkg2.pkg3.pkg4.origin_module_in_pkg)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_package_with_from_and_normal_imports(self) -> None:
        code = dedent("""\
            from pkg import mod4
            import pkg.mod4
            print(pkg.mod4)
            print(mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("pkg") + 1, self.destination_pkg_root)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod5") is not None)
        expected = dedent("""\
            from destination_pkg_root.pkg import mod4
            import destination_pkg_root.pkg.mod4
            print(destination_pkg_root.pkg.mod4)
            print(mod4)""")
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_package_with_from_and_normal_imports2(self) -> None:
        code = dedent("""\
            import pkg.mod4
            from pkg import mod4
            print(pkg.mod4)
            print(mod4)""")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("pkg") + 1, self.destination_pkg_root)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("destination_pkg_root.pkg.mod5") is not None)
        expected = dedent("""\
            import destination_pkg_root.pkg.mod4
            from destination_pkg_root.pkg import mod4
            print(destination_pkg_root.pkg.mod4)
            print(mod4)""")
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_package_and_retaining_blank_lines(self) -> None:
        pkg2 = testutils.create_package(self.project, "pkg2", self.pkg)
        code = dedent('''\
            """Docstring followed by blank lines."""

            import pkg.origin_module_in_pkg

            from pkg import origin_module_in_pkg
            from x import y
            from y import z
            from a import b
            from b import c
            print(pkg.origin_module_in_pkg)
            print(origin_module_in_pkg)''')
        self.mod1.write(code)
        self._move(self.origin_module_in_pkg, None, pkg2)
        expected = dedent('''\
            """Docstring followed by blank lines."""

            import pkg.pkg2.origin_module_in_pkg

            from x import y
            from y import z
            from a import b
            from b import c
            from pkg.pkg2 import origin_module_in_pkg
            print(pkg.pkg2.origin_module_in_pkg)
            print(origin_module_in_pkg)''')
        self.assertEqual(expected, self.mod1.read())

    def test_moving_functions_to_imported_module(self) -> None:
        code = dedent("""\
            import destination_module
            def a_func():
                var = destination_module.a_var
        """)
        self.destination_module.write("a_var = 1\n")
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("a_func") + 1, self.destination_module)
        expected = dedent("""\
            def a_func():
                var = a_var
            a_var = 1
        """)
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_resources_using_move_module_refactoring(self) -> None:
        self.origin_module.write("a_var = 1")
        self.mod2.write(dedent("""\
            import origin_module
            my_var = origin_module.a_var
        """))
        resource = self.origin_module
        dest_resource = self.destination_pkg_root
        self._move(resource, None, dest_resource)
        expected = dedent("""\
            import destination_pkg_root.origin_module
            my_var = destination_pkg_root.origin_module.a_var
        """)
        self.assertEqual(expected, self.mod2.read())
        self.assertTrue(self.destination_pkg_root.get_child("origin_module.py") is not None)

    def test_moving_resources_using_move_module_for_packages(self) -> None:
        self.mod1.write(dedent("""\
            import pkg
            my_pkg = pkg"""))
        self._move(self.pkg, None, self.destination_pkg_root)
        expected = dedent("""\
            import destination_pkg_root.pkg
            my_pkg = destination_pkg_root.pkg""")
        self.assertEqual(expected, self.mod1.read())
        self.assertTrue(self.destination_pkg_root.get_child("pkg") is not None)

    def test_moving_resources_using_move_module_for_init_dot_py(self) -> None:
        self.mod1.write(dedent("""\
            import pkg
            my_pkg = pkg"""))
        init = self.pkg.get_child("__init__.py")
        self._move(init, None, self.destination_pkg_root)
        self.assertEqual(
            dedent("""\
                import destination_pkg_root.pkg
                my_pkg = destination_pkg_root.pkg"""),
            self.mod1.read(),
        )
        self.assertTrue(self.destination_pkg_root.get_child("pkg") is not None)

    def test_moving_module_and_star_imports(self) -> None:
        self.origin_module.write("a_var = 1")
        self.mod2.write(dedent("""\
            from origin_module import *
            a = a_var
        """))
        self._move(self.origin_module, None, self.destination_pkg_root)
        self.assertEqual(
            dedent("""\
                from destination_pkg_root.origin_module import *
                a = a_var
            """),
            self.mod2.read(),
        )

    def test_moving_module_and_not_removing_blanks_after_imports(self) -> None:
        self.origin_module_in_pkg.write("a_var = 1")
        self.mod2.write(dedent("""\
            from pkg import origin_module_in_pkg
            import os


            print(origin_module_in_pkg.a_var)
        """))
        self._move(self.origin_module_in_pkg, None, self.project.root)
        self.assertEqual(
            dedent("""\
                import os
                import origin_module_in_pkg


                print(origin_module_in_pkg.a_var)
            """),
            self.mod2.read(),
        )

    def test_moving_module_refactoring_and_nonexistent_destinations(self) -> None:
        self.origin_module_in_pkg.write("a_var = 1")
        self.mod2.write(dedent("""\
            from pkg import origin_module_in_pkg
            import os


            print(origin_module_in_pkg.a_var)
        """))
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Move destination for modules should be packages.",
        ):
            self._move(self.origin_module_in_pkg, None, None)  # type: ignore[arg-type]

    def test_moving_methods_choosing_the_correct_class(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self):
                    pass
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertTrue(isinstance(mover, move.MoveMethod))

    def test_moving_methods_getting_new_method_for_empty_methods(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self):
                    pass
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    pass
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_for_constant_methods(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    return 1
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_passing_simple_parameters(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self, p):
                    return p
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, p):
                    return p
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_using_main_object(self) -> None:
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(host):
                    return host.attr
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, host):
                    return host.attr
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_renaming_main_object(self) -> None:
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self):
                    return self.attr
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, host):
                    return host.attr
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_gettin_new_method_with_keyword_arguments(self) -> None:
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self, p=None):
                    return p
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, p=None):
                    return p
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_gettin_new_method_with_many_kinds_arguments(self) -> None:
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self, p1, *args, **kwds):
                    return self.attr
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        expected = dedent("""\
            def new_method(self, host, p1, *args, **kwds):
                return host.attr
        """)
        self.assertEqual(expected, mover.get_new_method("new_method"))

    def test_moving_methods_getting_new_method_for_multi_line_methods(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self):
                    a = 2
                    return a
        """)
        self.origin_module.write(code)
        mover = move.create_move(self.project, self.origin_module, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    a = 2
                    return a
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_old_method_for_constant_methods(self) -> None:
        self.mod2.write(dedent("""\
            class B(object):
                pass
        """))
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        expected = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return self.attr.new_method()
        """)
        self.assertEqual(expected, self.origin_module.read())

    def test_moving_methods_getting_getting_changes_for_goal_class(self) -> None:
        self.mod2.write(dedent("""\
            class B(object):
                var = 1
        """))
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        expected = dedent("""\
            class B(object):
                var = 1


                def new_method(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_methods_getting_getting_changes_for_goal_class2(self) -> None:
        code = dedent("""\
            class B(object):
                var = 1

            class A(object):
                attr = B()
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        self.assertEqual(
            dedent("""\
                class B(object):
                    var = 1


                    def new_method(self):
                        return 1

                class A(object):
                    attr = B()
                    def a_method(self):
                        return self.attr.new_method()
            """),
            self.origin_module.read(),
        )

    def test_moving_methods_and_nonexistent_attributes(self) -> None:
        code = dedent("""\
            class A(object):
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Destination attribute <x> not found",
        ):
            self._move_to_attr(self.origin_module, code.index("a_method"), "x", new_name="new_method")

    def test_unknown_attribute_type(self) -> None:
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Unknown class type for attribute <attr>",
        ):
            self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")

    def test_moving_methods_and_moving_used_imports(self) -> None:
        self.mod2.write(dedent("""\
            class B(object):
                var = 1
        """))
        code = dedent("""\
            import sys
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return sys.version
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        code = dedent("""\
            import sys
            class B(object):
                var = 1


                def new_method(self):
                    return sys.version
        """)
        self.assertEqual(code, self.mod2.read())

    def test_moving_methods_getting_getting_changes_for_goal_class3(self) -> None:
        self.mod2.write(dedent("""\
            class B(object):
                pass
        """))
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        expected = dedent("""\
            class B(object):

                def new_method(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_methods_and_source_class_with_parameters(self) -> None:
        self.mod2.write(dedent("""\
            class B(object):
                pass
        """))
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self, p):
                    return p
        """)
        self.origin_module.write(code)
        self._move_to_attr(self.origin_module, code.index("a_method"), "attr", new_name="new_method")
        expected1 = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self, p):
                    return self.attr.new_method(p)
        """)
        self.assertEqual(expected1, self.origin_module.read())
        expected2 = dedent("""\
            class B(object):

                def new_method(self, p):
                    return p
        """)
        self.assertEqual(expected2, self.mod2.read())

    def test_moving_globals_to_a_module_with_only_docstrings(self) -> None:
        self.origin_module.write(dedent("""\
            import sys


            def f():
                print(sys.version)
        """))
        self.destination_module.write(dedent('''\
            """doc

            More docs ...

            """
        '''))
        self._move(self.origin_module, self.origin_module.read().index("f()") + 1, self.destination_module)
        self.assertEqual(
            dedent('''\
                """doc

                More docs ...

                """
                import sys


                def f():
                    print(sys.version)
            '''),
            self.destination_module.read(),
        )

    def test_moving_globals_to_a_module_with_only_docstrings2(self) -> None:
        code = dedent("""\
            import os
            import sys


            def f():
                print(sys.version, os.path)
        """)
        self.origin_module.write(code)
        self.destination_module.write(dedent('''\
            """doc

            More docs ...

            """
        '''))
        self._move(self.origin_module, self.origin_module.read().index("f()") + 1, self.destination_module)
        expected = dedent('''\
            """doc

            More docs ...

            """
            import os
            import sys


            def f():
                print(sys.version, os.path)
        ''')
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_a_global_when_it_is_used_after_a_multiline_str(self) -> None:
        code = dedent('''\
            def f():
                pass
            s = """\\
            """
            r = f()
        ''')
        self.origin_module.write(code)
        self._move(self.origin_module, code.index("f()") + 1, self.destination_module)
        expected = dedent('''\
            import destination_module
            s = """\\
            """
            r = destination_module.f()
        ''')
        self.assertEqual(expected, self.origin_module.read())

    def test_raising_an_exception_when_moving_non_package_folders(self) -> None:
        dir = self.project.root.create_folder("dir")
        with self.assertRaisesRegex(
            exceptions.RefactoringError,
            r"Cannot move non-package folder\.",
        ):
            move.create_move(self.project, dir)

    def test_moving_to_a_module_with_encoding_cookie(self) -> None:
        code = dedent("""\
            def f(): pass
        """)
        self.origin_module.write(code)
        self.destination_module.write("# -*- coding: utf-8 -*-")
        self._move(self.origin_module, code.index("f()") + 1, self.destination_module)
        expected = dedent("""\
            # -*- coding: utf-8 -*-
            def f(): pass
        """)
        self.assertEqual(expected, self.destination_module.read())

    def test_moving_decorated_function(self) -> None:
        self.origin_module.write(dedent("""\
            def hello(func):
                return func
            @hello
            def foo():
                pass
        """))
        self._move(self.origin_module, self.origin_module.read().index("foo") + 1, self.destination_module)
        self.assertEqual(
            dedent("""\
                def hello(func):
                    return func
            """), 
            self.origin_module.read(),
        )
        self.assertEqual(
            dedent("""\
                from origin_module import hello


                @hello
                def foo():
                    pass
            """),
            self.destination_module.read(),
        )

    def test_moving_decorated_class(self) -> None:
        self.origin_module.write(dedent("""\
            from dataclasses import dataclass
            @dataclass
            class AClass:
                pass
        """))
        self._move(self.origin_module, self.origin_module.read().index("AClass") + 1, self.destination_module)
        self.assertEqual("", self.origin_module.read())
        self.assertEqual(
            dedent("""\
                from dataclasses import dataclass


                @dataclass
                class AClass:
                    pass
            """),
            self.destination_module.read(),
        )



import os
import pathlib
import sys
from subprocess import check_call
from venv import EnvBuilder

import pytest

from rope.base import resources
from ropetest import testutils


@pytest.fixture(scope="session")
def session_venv(tmpdir_factory):
    path = tmpdir_factory.mktemp("venv")
    venv_path = pathlib.Path(path)

    builder = EnvBuilder(with_pip=True)
    builder.create(venv_path)

    yield venv_path


@pytest.fixture(scope="session")
def session_venv_pyvenv_cfg(session_venv):
    cfg = session_venv / "pyvenv.cfg"
    return dict(line.split(" = ") for line in cfg.read_text().splitlines())


@pytest.fixture(scope="session")
def session_venv_site_packages(session_venv, session_venv_pyvenv_cfg):
    if os.name == 'nt':
        return session_venv / f"Lib/site-packages"
    else:
        major, minor, patch = session_venv_pyvenv_cfg["version"].split(".")
        return session_venv / f"lib/python{major}.{minor}/site-packages"


@pytest.fixture(scope='session')
def session_venv_python_executable(session_venv):
    # Get the path to the Python executable inside the venv
    if os.name == 'nt':
        python_executable = session_venv / 'Scripts' / 'python.exe'
    else:
        python_executable = session_venv / 'bin' / 'python'

    # Yield the Python executable path
    yield python_executable


@pytest.fixture
def project(session_venv, session_venv_site_packages):
    project = testutils.sample_project(
        python_path=[str(session_venv_site_packages)],
    )
    yield project
    testutils.remove_project(project)


@pytest.fixture
def project_path(project):
    yield pathlib.Path(project.address)


@pytest.fixture
def project2(session_venv):
    project = testutils.sample_project(
        "sample_project2",
        python_path=[str(session_venv_site_packages)],
    )
    yield project
    testutils.remove_project(project)


"""
Standard project structure for pytest fixtures
/mod1.py            -- mod1
/pkg1/__init__.py   -- pkg1
/pkg1/mod2.py       -- mod2
"""

@pytest.fixture
def mod1(project) -> resources.File:
    return testutils.create_module(project, "mod1")


@pytest.fixture
def pkg1(project) -> resources.Folder:
    return testutils.create_package(project, "pkg1")


@pytest.fixture
def mod2(project, pkg1) -> resources.Folder:
    return testutils.create_module(project, "mod2", pkg1)


@pytest.fixture(scope="session")
def external_fixturepkg(session_venv, session_venv_python_executable):
    check_call([
        session_venv_python_executable,
        "-m",
        "pip",
        "install",
        "--force-reinstall",
        "ropetest-package-fixtures/external_fixturepkg/dist/external_fixturepkg-1.0.0-py3-none-any.whl",
    ])
    yield
    check_call([sys.executable, "-m", "pip", "uninstall", "--yes", "external-fixturepkg"])



"""AutoImport module for rope."""

from __future__ import annotations

import contextlib
import json
import re
import secrets
import sqlite3
import sys
import warnings
from collections import OrderedDict
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from datetime import datetime
from hashlib import sha256
from itertools import chain
from pathlib import Path
from threading import local
from typing import (
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TYPE_CHECKING,
)

from rope.base import exceptions, libutils, resourceobserver, taskhandle, versioning
from rope.base.project import Project
from rope.base.resources import Resource
from rope.contrib.autoimport import models
from rope.contrib.autoimport.defs import (
    Alias,
    ModuleFile,
    Name,
    NameType,
    Package,
    PackageType,
    SearchResult,
    Source,
)
from rope.contrib.autoimport.parse import get_names
from rope.contrib.autoimport.utils import (
    get_files,
    get_modname_from_path,
    get_package_tuple,
    sort_and_deduplicate,
    sort_and_deduplicate_tuple,
)
from rope.refactor import importutils


if TYPE_CHECKING:
    from collections.abc import Collection


def get_future_names(
    packages: List[Package], underlined: bool, job_set: taskhandle.BaseJobSet
) -> Generator[Future[Collection[Name]], None, None]:
    """Get all names as futures."""
    with ProcessPoolExecutor() as executor:
        for package in packages:
            for module in get_files(package, underlined):
                job_set.started_job(module.modname)
                job_set.increment()
                yield executor.submit(get_names, module, package)


def filter_packages(
    packages: Iterable[Package], underlined: bool, existing: List[str]
) -> Iterable[Package]:
    """Filter list of packages to parse."""
    if underlined:

        def filter_package(package: Package) -> bool:
            return package.name not in existing

    else:

        def filter_package(package: Package) -> bool:
            return package.name not in existing and not package.name.startswith("_")

    return filter(filter_package, packages)


_deprecated_default: bool = object()  # type: ignore


class AutoImport:
    """A class for finding the module that provides a name.

    This class maintains a cache of global names in python modules.
    Note that this cache is not accurate and might be out of date.

    """

    memory: bool
    project: Project
    project_package: Package
    underlined: bool

    def __init__(
        self,
        project: Project,
        observe: bool = True,
        underlined: bool = False,
        memory: bool = _deprecated_default,
    ):
        """Construct an AutoImport object.

        Parameters
        ___________
        project : rope.base.project.Project
            the project to use for project imports
        observe : bool
            if true, listen for project changes and update the cache.
        underlined : bool
            If `underlined` is `True`, underlined names are cached, too.
        memory:
            If true, don't persist to disk

            DEPRECATION NOTICE: The default value will change to use an on-disk
            database by default in the future. If you want to use an in-memory
            database, you need to pass `memory=True` explicitly:

                autoimport = AutoImport(..., memory=True)
        """
        self.project = project
        project_package = get_package_tuple(project.root.pathlib, project)
        assert project_package is not None
        assert project_package.path is not None
        self.project_package = project_package
        self.underlined = underlined
        self.memory = memory
        if memory is _deprecated_default:
            self.memory = True
            warnings.warn(
                "The default value for `AutoImport(memory)` argument will "
                "change to use an on-disk database by default in the future. "
                "If you want to use an in-memory database, you need to pass "
                "`AutoImport(memory=True)` explicitly.",
                DeprecationWarning,
            )
        self.thread_local = local()
        self.connection = self.create_database_connection(
            project=project,
            memory=memory,
        )
        self._setup_db()
        if observe:
            observer = resourceobserver.ResourceObserver(
                changed=self._changed, moved=self._moved, removed=self._removed
            )
            project.add_observer(observer)

    @classmethod
    def create_database_connection(
        cls,
        *,
        project: Optional[Project] = None,
        memory: bool = False,
    ) -> sqlite3.Connection:
        """
        Create an sqlite3 connection

        project : rope.base.project.Project
            the project to use for project imports
        memory : bool
            if true, don't persist to disk
        """

        def calculate_project_hash(data: str) -> str:
            return sha256(data.encode()).hexdigest()

        if not memory and project is None:
            raise Exception("if memory=False, project must be provided")
        if memory or project is None or project.ropefolder is None:
            # Allows the in-memory db to be shared across threads
            # See https://www.sqlite.org/inmemorydb.html
            project_hash: str
            if project is None:
                project_hash = secrets.token_hex()
            elif project.ropefolder is None:
                project_hash = calculate_project_hash(project.address)
            else:
                project_hash = calculate_project_hash(project.ropefolder.real_path)
            return sqlite3.connect(
                f"file:rope-{project_hash}:?mode=memory&cache=shared", uri=True
            )
        else:
            return sqlite3.connect(project.ropefolder.pathlib / "autoimport.db")

    @property
    def connection(self) -> sqlite3.Connection:
        """
        Creates a new connection if called from a new thread.

        This makes sure AutoImport can be shared across threads.
        """
        if not hasattr(self.thread_local, "connection"):
            self.thread_local.connection = self.create_database_connection(
                project=self.project,
                memory=self.memory,
            )
        return self.thread_local.connection

    @connection.setter
    def connection(self, value: sqlite3.Connection):
        self.thread_local.connection = value

    def _setup_db(self):
        models.Metadata.create_table(self.connection)
        version_hash = list(
            self._execute(models.Metadata.objects.select("version_hash"))
        )
        current_version_hash = versioning.calculate_version_hash(self.project)
        if not version_hash or version_hash[0][0] != current_version_hash:
            self.clear_cache()

    def import_assist(self, starting: str):
        """
        Find modules that have a global name that starts with `starting`.

        For a more complete list, use the search or search_full methods.

        Parameters
        __________
        starting : str
            what all the names should start with
        Return
        __________
        Return a list of ``(name, module)`` tuples
        """
        results = self._execute(
            models.Name.import_assist.select("name", "module", "source"), (starting,)
        ).fetchall()
        return sort_and_deduplicate_tuple(
            results
        )  # Remove duplicates from multiple occurrences of the same item

    def search(self, name: str, exact_match: bool = False) -> List[Tuple[str, str]]:
        """
        Search both modules and names for an import string.

        This is a simple wrapper around search_full with basic sorting based on Source.

        Returns a sorted list of import statement, modname pairs
        """
        results: List[Tuple[str, str, int]] = [
            (statement, import_name, source)
            for statement, import_name, source, type in self.search_full(
                name, exact_match
            )
        ]
        return sort_and_deduplicate_tuple(results)

    def search_full(
        self,
        name: str,
        exact_match: bool = False,
        ignored_names: Optional[Set[str]] = None,
    ) -> Generator[SearchResult, None, None]:
        """
        Search both modules and names for an import string.

        Parameters
        __________
        name: str
            Name to search for
        exact_match: bool
            If using exact_match, only search for that name.
            Otherwise, search for any name starting with that name.
        ignored_names : Set[str]
            Will ignore any names in this set

        Return
        __________
        Unsorted Generator of SearchResults. Each is guaranteed to be unique.
        """
        if ignored_names is None:
            ignored_names = set()
        results = set(self._search_name(name, exact_match))
        results = results.union(self._search_module(name, exact_match))
        for result in results:
            if result.name not in ignored_names:
                yield result

    def _search_name(
        self, name: str, exact_match: bool = False
    ) -> Generator[SearchResult, None, None]:
        """
        Search both names for available imports.

        Returns the import statement, import name, source, and type.
        """
        if not exact_match:
            name = name + "%"  # Makes the query a starts_with query
        for import_name, module, source, name_type in self._execute(
            models.Name.search_by_name_like.select("name", "module", "source", "type"),
            (name,),
        ):
            yield (
                SearchResult(
                    f"from {module} import {import_name}",
                    import_name,
                    source,
                    name_type,
                )
            )

    def _search_module(
        self, name: str, exact_match: bool = False
    ) -> Generator[SearchResult, None, None]:
        """
        Search both modules for available imports.

        Returns the import statement, import name, source, and type.
        """
        if not exact_match:
            name = name + "%"  # Makes the query a starts_with query
        for module, source in self._execute(
            models.Name.search_submodule_like.select("module", "source"), (name,)
        ):
            parts = module.split(".")
            import_name = parts[-1]
            remaining = parts[0]
            for part in parts[1:-1]:
                remaining += "."
                remaining += part
            yield (
                SearchResult(
                    f"from {remaining} import {import_name}",
                    import_name,
                    source,
                    NameType.Module.value,
                )
            )
        for module, source in self._execute(
            models.Name.search_module_like.select("module", "source"), (name,)
        ):
            if "." in module:
                continue
            yield SearchResult(
                f"import {module}", module, source, NameType.Module.value
            )
        for alias, module, source in self._execute(
            models.Alias.search_modules_with_alias.select("alias", "module", "source"),
            (name,),
        ):
            yield SearchResult(
                f"import {module} as {alias}", alias, source, NameType.Module.value
            )

    def get_modules(self, name) -> List[str]:
        """Get the list of modules that have global `name`."""
        results = self._execute(
            models.Name.search_by_name.select("module", "source"), (name,)
        ).fetchall()
        return sort_and_deduplicate(results)

    def get_all_names(self) -> List[str]:
        """Get the list of all cached global names."""
        return self._execute(models.Name.objects.select("name")).fetchall()

    def _dump_all(self) -> Tuple[List[Name], List[Package]]:
        """Dump the entire database."""
        name_results = self._execute(models.Name.objects.select_star()).fetchall()
        package_results = self._execute(models.Package.objects.select_star()).fetchall()
        return name_results, package_results

    def generate_cache(
        self,
        resources: Optional[List[Resource]] = None,
        underlined: bool = False,
        task_handle: taskhandle.BaseTaskHandle = taskhandle.DEFAULT_TASK_HANDLE,
    ):
        """Generate global name cache for project files.

        If `resources` is a list of `rope.base.resource.File`, only
        those files are searched; otherwise all python modules in the
        project are cached.
        """
        if resources is None:
            resources = self.project.get_python_files()
        job_set = task_handle.create_jobset(
            "Generating autoimport cache", len(resources)
        )
        self._execute(
            models.Package.delete_by_package_name, (self.project_package.name,)
        )
        futures = []
        with ProcessPoolExecutor() as executor:
            for file in resources:
                job_set.started_job(f"Working on {file.path}")
                module = self._resource_to_module(file, underlined)
                futures.append(executor.submit(get_names, module, self.project_package))
        for future in as_completed(futures):
            self._add_names(future.result())
            job_set.finished_job()
        self.connection.commit()

    def generate_modules_cache(
        self,
        modules: Optional[List[str]] = None,
        task_handle: taskhandle.BaseTaskHandle = taskhandle.DEFAULT_TASK_HANDLE,
        single_thread: bool = False,
        underlined: Optional[bool] = None,
    ):
        """
        Generate global name cache for external modules listed in `modules`.

        If no modules are provided, it will generate a cache for every module available.
        This method searches in your sys.path and configured python folders.
        Do not use this for generating your own project's internal names,
        use generate_resource_cache for that instead.
        """
        underlined = self.underlined if underlined is None else underlined

        packages: List[Package] = (
            self._get_available_packages()
            if modules is None
            else list(self._get_packages_from_modules(modules))
        )

        existing = self._get_packages_from_cache()
        packages = list(filter_packages(packages, underlined, existing))
        if not packages:
            return
        self._add_packages(packages)
        job_set = task_handle.create_jobset("Generating autoimport cache", 0)
        if single_thread:
            for package in packages:
                for module in get_files(package, underlined):
                    job_set.started_job(module.modname)
                    for name in get_names(module, package):
                        self._add_name(name)
                        job_set.finished_job()
        else:
            for future_name in as_completed(
                get_future_names(packages, underlined, job_set)
            ):
                self._add_names(future_name.result())
                job_set.finished_job()

        self.connection.commit()

    def _get_packages_from_modules(self, modules: List[str]) -> Iterator[Package]:
        for modname in modules:
            package = self._find_package_path(modname)
            if package is None:
                continue
            yield package

    def update_module(self, module: str):
        """Update a module in the cache, or add it if it doesn't exist."""
        self._del_if_exist(module)
        self.generate_modules_cache([module])

    def close(self):
        """Close the autoimport database."""
        self.connection.commit()
        self.connection.close()

    def get_name_locations(self, name):
        """Return a list of ``(resource, lineno)`` tuples."""
        result = []
        modules = self._execute(
            models.Name.search_by_name_like.select("module"), (name,)
        ).fetchall()
        for module in modules:
            with contextlib.suppress(exceptions.ModuleNotFoundError):
                module_name = module[0]
                if module_name.startswith(f"{self.project_package.name}."):
                    module_name = ".".join(module_name.split("."))
                pymodule = self.project.get_module(module_name)
                if name in pymodule:
                    pyname = pymodule[name]
                    module, lineno = pyname.get_definition_location()
                    if module is not None:
                        resource = module.get_module().get_resource()
                        if resource is not None and lineno is not None:
                            result.append((resource, lineno))
        return result

    def clear_cache(self):
        """Clear all entries in global-name cache.

        It might be a good idea to use this function before
        regenerating global names.

        """
        with self.connection:
            self._execute(models.Name.objects.drop_table())
            self._execute(models.Alias.objects.drop_table())
            self._execute(models.Package.objects.drop_table())
            self._execute(models.Metadata.objects.drop_table())
            models.Name.create_table(self.connection)
            models.Alias.create_table(self.connection)
            models.Package.create_table(self.connection)
            models.Metadata.create_table(self.connection)
            self.add_aliases(self.project.prefs.autoimport.aliases)
            data = (
                versioning.calculate_version_hash(self.project),
                json.dumps(versioning.get_version_hash_data(self.project)),
                datetime.utcnow().isoformat(),
            )
            assert models.Metadata.columns == [
                "version_hash",
                "hash_data",
                "created_at",
            ]
            self._execute(models.Metadata.objects.insert_into(), data)

            self.connection.commit()

    def find_insertion_line(self, code):
        """Guess at what line the new import should be inserted."""
        match = re.search(r"^(def|class)\s+", code)
        if match is not None:
            code = code[: match.start()]
        try:
            pymodule = libutils.get_string_module(self.project, code)
        except exceptions.ModuleSyntaxError:
            return 1
        testmodname = "__rope_testmodule_rope"
        importinfo = importutils.NormalImport(((testmodname, None),))
        module_imports = importutils.get_module_imports(self.project, pymodule)
        module_imports.add_import(importinfo)
        code = module_imports.get_changed_source()
        offset = code.index(testmodname)
        lineno = code.count("\n", 0, offset) + 1
        return lineno

    def update_resource(
        self, resource: Resource, underlined: bool = False, commit: bool = True
    ):
        """Update the cache for global names in `resource`."""
        underlined = underlined if underlined else self.underlined
        module = self._resource_to_module(resource, underlined)
        self._del_if_exist(module_name=module.modname, commit=False)
        for name in get_names(module, self.project_package):
            self._add_name(name)
        if commit:
            self.connection.commit()

    def _changed(self, resource):
        if not resource.is_folder():
            self.update_resource(resource)

    def _moved(self, resource: Resource, newresource: Resource):
        if not resource.is_folder():
            modname = self._resource_to_module(resource).modname
            self._del_if_exist(modname)
            self.update_resource(newresource)

    def _del_if_exist(self, module_name, commit: bool = True):
        self._execute(models.Name.delete_by_module_name, (module_name,))
        if commit:
            self.connection.commit()

    def _get_python_folders(self) -> List[Path]:
        def filter_folders(folder: Path) -> bool:
            return folder.is_dir() and folder.as_posix() != "/usr/bin"

        folders = self.project.get_python_path_folders()
        folder_paths = filter(filter_folders, map(Path, folders))
        return list(OrderedDict.fromkeys(folder_paths))

    def _safe_iterdir(self, folder: Path):
        try:
            dirs = folder.iterdir()
            while True:
                try:
                    yield next(dirs)
                except PermissionError:
                    pass
                except StopIteration:
                    break
        except PermissionError:
            pass

    def _get_available_packages(self) -> List[Package]:
        packages: List[Package] = [
            Package(module, Source.BUILTIN, None, PackageType.BUILTIN)
            for module in sys.builtin_module_names
        ]
        for folder in self._get_python_folders():
            for package in self._safe_iterdir(folder):
                package_tuple = get_package_tuple(package, self.project)
                if package_tuple is None:
                    continue
                packages.append(package_tuple)
        return packages

    def _add_packages(self, packages: List[Package]):
        data = [(p.name, str(p.path)) for p in packages]
        self._executemany(models.Package.objects.insert_into(), data)

    def _get_packages_from_cache(self) -> List[str]:
        existing: List[str] = list(
            chain(*self._execute(models.Package.objects.select_star()).fetchall())
        )
        existing.append(self.project_package.name)
        return existing

    def _removed(self, resource):
        if not resource.is_folder():
            modname = self._resource_to_module(resource).modname
            self._del_if_exist(modname)

    @staticmethod
    def _convert_name(name: Name) -> tuple:
        return (
            name.name,
            name.modname,
            name.package,
            name.source.value,
            name.name_type.value,
        )

    def add_aliases(self, aliases: Collection[Alias]):
        if aliases:
            self._executemany(models.Alias.objects.insert_into(), aliases)

    def _add_names(self, names: Collection[Name]):
        if names:
            self._executemany(
                models.Name.objects.insert_into(),
                [self._convert_name(name) for name in names],
            )

    def _add_name(self, name: Name):
        self._execute(models.Name.objects.insert_into(), self._convert_name(name))

    def _find_package_path(self, target_name: str) -> Optional[Package]:
        if target_name in sys.builtin_module_names:
            return Package(target_name, Source.BUILTIN, None, PackageType.BUILTIN)
        for folder in self._get_python_folders():
            for package in self._safe_iterdir(folder):
                package_tuple = get_package_tuple(package, self.project)
                if package_tuple is None:
                    continue
                name, source, package_path, package_type = package_tuple
                if name == target_name:
                    return package_tuple

        return None

    def _resource_to_module(
        self, resource: Resource, underlined: bool = False
    ) -> ModuleFile:
        assert self.project_package.path
        underlined = underlined if underlined else self.underlined
        resource_path: Path = resource.pathlib
        # The project doesn't need its name added to the path,
        # since the standard python file layout accounts for that
        # so we set add_package_name to False
        resource_modname: str = get_modname_from_path(
            resource_path, self.project_package.path, add_package_name=False
        )
        return ModuleFile(
            resource_path,
            resource_modname,
            underlined,
            resource_path.name == "__init__.py",
        )

    def _execute(self, query: models.FinalQuery, *args, **kwargs):
        assert isinstance(query, models.FinalQuery)
        return self.connection.execute(query._query, *args, **kwargs)

    def _executemany(self, query: models.FinalQuery, *args, **kwargs):
        assert isinstance(query, models.FinalQuery)
        return self.connection.executemany(query._query, *args, **kwargs)
